#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#include <strings.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>

#define SERVER_PORT 7111
#define LEN_RECVBUFFER 1000


static int sfd;

/*--------------------------------------------------------------------*/

//Signalhandler zum Abbrechen des Servers
static void sighandler(int sig)
{

  switch( sig ) {

  case SIGPIPE: // Nur den zugehörigen Client abbrechen
    break;

  default:
    printf("Signal %d caught\n", sig);

    if(0 > close(sfd))
      {
        perror("Problem with close()");
        exit(1);
      }

    exit(0);
    break;
  }


}

/*--------------------------------------------------------------------*/

// Thread function:
// Eingabeparameter ist 1 Pointer.
// Falls mehrere Variablen uebergeben werden sollen,
// sird eine Struktur verwendet, auf die der Pointer zeigt.
static void *client_handler(void *p) {

  int clientsfd = *(int *)p;
  char *recvbuffer = NULL;
  ssize_t lenrecv, lensent, len;

  // den dynamisch angelegten Datenbereich von p
  // wieder freigeben
  free(p);

  recvbuffer = (char *)calloc(1, LEN_RECVBUFFER+1);
  if (!recvbuffer) {
    perror("Problem with malloc()");
    goto out;
  }

  fcntl(clientsfd,O_NONBLOCK);

  while ( 1 ) {

    //lesen
    lenrecv = read(clientsfd, recvbuffer, LEN_RECVBUFFER);
    if ( lenrecv < 0) {
      perror("problem with read()");
      goto out;
    }
    if ( 0 < lenrecv ) {
      recvbuffer[lenrecv] = 0;
      printf("Server-Thread %d received %s\n",clientsfd,recvbuffer);

      //wenn STOP empfangen wird, beendet sich der Client-Thread
      if ( !strncmp(recvbuffer,"STOP",strlen("STOP")) ) {
        printf("Server-Child exits\n");
        goto out;
      }

      // wir moechten lenrcv Bytes ueber Socket versenden
      // Dazu machen wir eine Schleife, bis tatsaechlich
      // alle Bytes vom BS akzeptiert wurden.
      // Bei jedem Write() wird die Anfangsadresse des
      // Sendepuffers auf das ERSTE NOCH NICHT VERSENDETE
      // Byte gesetzt.
      lensent = 0;
      while ( lensent < lenrecv ) {
        if ( 0 > (len = write(clientsfd,
                              recvbuffer + lensent,
                              lenrecv - lensent)) ) {
          perror("write()-call");
          goto out;
        }
        lensent = lensent + len;
      }
    }
  }
 out:
  free(recvbuffer);
  close(clientsfd);
  return NULL;
}

/*--------------------------------------------------------------------*/
int main() {
  int clientsfd;
  struct  sockaddr_in saddr;
  struct sockaddr clientaddr;
  socklen_t lenclientaddr;
  int optval;

  pthread_t th;
  int *pPtr;

  //Signalhandler zum Beenden
  signal(SIGTERM, sighandler);
  signal(SIGPIPE, sighandler);
  signal(SIGINT, sighandler);

  /*--- PF_INET waehlt die IP-Protokollfamilie aus.
    SOCK_STREAM wahlt Byte-Strom-orientierte Kommunikation aus.
    Das bedeutet zZt automatisch TCP ueber IP ---*/
  if ( 0 > ( sfd = socket(PF_INET,SOCK_STREAM,0) ) ) {
    perror("socket()-call");
    return 1;
  }

  /*--- das BS darf die IP/PORT-Adresse sofort
    wieder vergeben: ---*/
  optval = 1;
  if ( -1 == setsockopt(sfd,SOL_SOCKET,SO_REUSEADDR,
                        &optval,sizeof(optval)) ) {
    perror("setsockopt()-call");
  }

  /* Server bindet IP-Adresse und Port an den Socket ---*/
  memset(&saddr, 0, sizeof(saddr)); // saddr initialisieren
  saddr.sin_family = AF_INET; // Protokollfamilie
  saddr.sin_addr.s_addr = htonl(INADDR_ANY); // trage
  // IP Adresse des lokalen Rechners ein.
  //htonl(): "Host to Network LONG" (32-Bit-Worte):
  // IP und Port-Adressen muessen in der NETWORK
  // BYTE ORDER (Big Endian Format) angegeben werden.
  // htons(): Wandlung von 16-Bit Worten in Network
  // Byte Order.
  saddr.sin_port = htons(SERVER_PORT);

  if ( 0 > bind(sfd,(struct sockaddr *)&saddr, sizeof(saddr)) ) {
    perror("bind()-call");
    return 1;
  }

  // Server teilt dem BS mit, dass er nun bereit ist,
  // auf dem Socket zu hoeren. "5" ist das BACKLOG, dh, die
  // Anzahl der Verbindungsaufbau-Wuensche, die das BS
  // intern puffert.
  if ( 0 > listen(sfd,5) ) {
    perror("listen()-call");
    return 1;
  }

  // Standardverfahren fuer das Entwurfsmuster PARALELL SERVER:
  // Fuer jeden aktiven Client wird vom Server ein Kindprozess
  // erzeugt, der diesen Client bedient.
  while (1) {

    lenclientaddr = sizeof(clientaddr);
    // accept() liefert einen NEUEN Socket Filedescriptor
    // zurueck, mit dem die Kommunikation zum CLient,
    // der das zugehoerige connect() ausgeloest hat
    // abgewickelt wird.
    // &clientaddr gibt die Adresse des Puffers an,
    // in die das BS die Adresse des Clients schreiben soll.
    // Bei lenclientaddr muss VOR dem Aufruf die
    // maximale Laenge von clientaddr stehen. NACH dem
    // Aufruf steht in lenclientaddr die tatsaechliche
    // Laenge der empfangenen Adresse.
    clientsfd = accept(sfd,&clientaddr,&lenclientaddr);
    if ( clientsfd < 0 ) {
      perror("accept()-call");
      return 1;
    }

    printf("Clientsfd %d\n", clientsfd);

    // Hier wird ein thread  erzeugt:
    pPtr = (int *)malloc(sizeof(int));
    if (!pPtr) {
      perror("Problem with malloc()");
      return 1;
    }
    *pPtr = clientsfd;

    

    // th: variable fuer Thread Identifikation
    // 2. Parameter: Auswahl der Scheduling
    // Policy; SCHED_FIFO, SCHED_RR oder SCHED_OTHER
    // 3. Parameter: Funktionspointer auf den Thread
    // statt pPtr kann man nicht einfach &clientsfd
    // uebergeben: Zum Zeitpunkt, zu dem der Thread
    // tatsaechlich aktiv, wird, kann in clientsfd
    // schon ein anderer Wert stehen.
    if ( pthread_create(&th, NULL, client_handler, pPtr) ) {
      perror("pthread_create()-call");
      return 1;
    }
  }

}




